
//	Copyright (c) 1982 Dyadic Systems Limited	

/*	x f a c c

	external function access routines
*/

#include <apl.h>
#include <support.h>
#include <string.h>
#include <math_iso99.h>
#include <decf.h>

static int32_t	roundi(double fval);
static aplfloat64 dgetz(void *vbase, BOUND indx);
static aplint32 lgetD(void *vbase, BOUND indx);
static aplcmpx128 zgetz(void *vbase, BOUND indx);
static apldecf128 DgetD(void *vbase, BOUND indx);

//	d o u b l e   p u t  &  g e t	

typedef	struct {
	UCH cfloat[sizeof(aplfloat64)];
} Doubarray;

typedef struct {
	union {
		Doubarray sfloat;
		aplfloat64 dfloat;
	} ufloat;
} Doubunion;

void putd(aplfloat64 val, aplfloat64 *base, BOUND indx)
{
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.dfloat = val;
	*((Doubarray*)(base+indx)) = dunion.ufloat.sfloat;
#else
	base[indx] = val;
#endif // DALIGN 
}

aplfloat64 getd(void *vbase, BOUND indx)
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return dunion.ufloat.dfloat;
#else

	return base[indx];

#endif // DALIGN 
}
//	c h a r  f r o m  a n y t h i n g  ( s u i t a b l e )	

static aplchar8 cgetc(void *vbase, BOUND indx)
{
	aplchar8 *base = vbase;

	return base[indx];
}

static aplchar8 cgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initchar(array))(array->shape, 0);
}

static aplchar8 cdomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	i n t   f r o m   a n y t h i n g	

static aplint16 igetb(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplint16 igets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return base[indx];
}

static aplint16 igeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplint16 igetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplint16 igetd(void *vbase, BOUND indx)	// rounds double to nearest int 
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return roundi(dunion.ufloat.dfloat);
#else
	return roundi(base[indx]);
#endif // DALIGN 
}

static aplint16 igetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initintg(array))(array->shape, 0);
}

static aplint16 igetz(void *vbase, BOUND indx)
{
	return roundi(dgetz(vbase, indx));
}

static aplint16 igetD(void *vbase, BOUND indx)
{
	aplint32 i32 = lgetD(vbase, indx);
	aplint16 i16 = i32;

	if(i16 != i32)
		domain();
	return i16;
}

static aplint16 idomx(void *vbase, BOUND indx)
	{
	domain();
	return 0; // keep compiler happy
	}

//	l o n g   f r o m   a n y t h i n g	

static aplint32 lgetb(void *vbase, BOUND indx)
{
	apluns8 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplint32 lgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return base[indx];
}

static aplint32 lgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplint32 lgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplint32 lgetd(void *vbase, BOUND indx)	// rounds double to nearest long
{
	aplfloat64 *base = vbase;
#ifdef	DALIGN
	Doubunion dunion;

	dunion.ufloat.sfloat = *((Doubarray*)(base+indx));
	return roundi(dunion.ufloat.dfloat);
#else
	return roundi(base[indx]);
#endif // DALIGN 
}

static aplint32 lgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initlong(array))(array->shape, 0);
}

static aplint32 lgetz(void *vbase, BOUND indx)
{
	return roundi(dgetz(vbase, indx));
}

static aplint32 lgetD(void *vbase, BOUND indx)
{
	apldecf128 D = DgetD(vbase, indx);
	if(decf_lt(D, DECF_ZERO))
		D = decf_add(D, DECF_HALF);
	else
		D = decf_sub(D, DECF_HALF);

	return decf_to_int32(D);
}

static aplint32 ldomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	d o u b l e   f r o m   a n y t h i n g	

static aplfloat64 dgetb(void *vbase, BOUND indx)
{
	apluns8 *base = vbase;

	return ((base[indx>>3] << (indx&7)) & 0x80 ) != 0;
}

static aplfloat64 dgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;

	return  (int)base[indx];
}

static aplfloat64 dgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;

	return base[indx];
}

static aplfloat64 dgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;

	return base[indx];
}

static aplfloat64 dgetd(void *vbase, BOUND indx)
{
	aplfloat64 *base = vbase;
	aplfloat64 d;

	memcpy(&d,  base + indx, sizeof(d));	// alignment ?
	return d;
}

static aplfloat64 dgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	register ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();
	return (*initdoub(array))(array->shape, 0);
}

static aplfloat64 dgetz(void *vbase, BOUND indx)
{
	aplcmpx128 z = zgetz(vbase, indx);

	if(z.im)
		domain();
	return z.re;
}

static aplfloat64 dgetD(void *vbase, BOUND indx)
{
	apldecf128 D = DgetD(vbase, indx);
	aplfloat64 d = decf_to_double(D);

	if(isinf(d))
		domain();
	return d;
}

static aplfloat64 ddomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

//	p n t r  f r o m  a n y t h i n g	

static ARRAY *pgetc(void *vbase, BOUND indx)
{
	aplchar8 *base = vbase;
	register ARRAY *array;

	array = mkarray(APLNCHAR, 0, 0);

	*(aplchar8 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetb(void *vbase, BOUND indx)
{
	register ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = lgetb(vbase,indx);

	return array;
}

static ARRAY *pgets(void *vbase, BOUND indx)
{
	aplint8 *base = vbase;
	ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgeti(void *vbase, BOUND indx)
{
	aplint16 *base = vbase;
	register ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetl(void *vbase, BOUND indx)
{
	aplint32 *base = vbase;
	ARRAY *array;

	array = mkarray(APLLONG, 0, 0);

	*(aplint32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetd(void *vbase, BOUND indx)
{
	aplfloat64 *base = vbase;
	ARRAY *array;

	array = mkarray(APLDOUB, 0, 0);

	*(aplfloat64 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;

	return base[indx];
}

static ARRAY *pdomx(void *vbase, BOUND indx)
	{
	domain();
	return 0;	// keep compiler happy
	}

static ARRAY *pgetW8(void *vbase, BOUND indx)
{
	aplWchar8 *base = vbase;
	ARRAY *array;

	array = mkarray(APLWCHAR8, 0, 0);
	*(aplWchar8 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetW16(void *vbase, BOUND indx)
{
	aplWchar16 *base = vbase;
	ARRAY *array;

	array = mkarray(APLWCHAR16, 0, 0);
	*(aplWchar16 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetW32(void *vbase, BOUND indx)
{
	aplWchar32 *base = vbase;
	ARRAY *array;

	array = mkarray(APLWCHAR32, 0, 0);
	*(aplWchar32 *)array->shape = base[indx];

	return array;
}

static ARRAY *pgetz(void *vbase, BOUND indx)
{
	aplcmpx128 *base = vbase;
	ARRAY *array;

	array = mkarray(APLCMPX, 0, 0);
	memcpy(array->shape, base + indx, sizeof(aplcmpx128)); // alignment ?

	return array;
}

static ARRAY *pgetD(void *vbase, BOUND indx)
{
	apldecf128 *base = vbase;
	ARRAY *array;

	array = mkarray(APLDECF, 0, 0);
	memcpy(array->shape, base + indx, sizeof(apldecf128)); // alignment ?

	return array;
}

static aplWchar8 W8domx(void *vbase, BOUND indx)
	{
	domain();
	return 0;
	}

static aplWchar8 W8getp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initWchar8(array))(array->shape, 0);
}


static aplWchar8 W8getW8(void *vbase, BOUND indx)
{
	aplWchar8 *base = vbase;

	return base[indx];
}

static aplchar8 W8getW16(void *vbase, BOUND indx)
{
	aplWchar16 *base = vbase;
	aplWchar16 w16 = base[indx];
	aplWchar8  w8  = w16;

	if(w8 != w16)
		domain();

	return w8;
}

static aplchar8 W8getW32(void *vbase, BOUND indx)
{
	aplWchar32 *base = vbase;
	aplWchar32 w32 = base[indx];
	aplWchar8  w8  = w32;

	if(w8 != w32)
		domain();

	return w8;
}

static aplWchar16 W16domx(void *vbase, BOUND indx)
	{
	domain();
	return 0;
	}

static aplWchar16 W16getp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initWchar16(array))(array->shape, 0);
}

static aplWchar16 W16getW8(void *vbase, BOUND indx)
{
	aplWchar8 *base = vbase;

	return base[indx];
}

static aplWchar16 W16getW16(void *vbase, BOUND indx)
{
	aplWchar16 *base = vbase;

	return base[indx];
}

static aplWchar16 W16getW32(void *vbase, BOUND indx)
{
	aplWchar32 *base = vbase;
	aplWchar32 w32 = base[indx];
	aplWchar16 w16  = w32;

	if(w16 != w32)
		domain();

	return w16;
}

static aplWchar32 W32domx(void *vbase, BOUND indx)
	{
	domain();
	return 0;
	}

static aplWchar32 W32getp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initWchar32(array))(array->shape, 0);
}

static aplWchar32 W32getW8(void *vbase, BOUND indx)
{
	aplWchar8 *base = vbase;

	return base[indx];
}

static aplWchar32 W32getW16(void *vbase, BOUND indx)
{
	aplWchar16 *base = vbase;

	return base[indx];
}

static aplcmpx128 zdomx(void *vbase, BOUND indx)
	{
	aplcmpx128 z = {0.0, 0.0};

	domain();
	return z;
	}

static aplWchar32 W32getW32(void *vbase, BOUND indx)
{
	aplWchar32 *base = vbase;

	return base[indx];
}

static aplcmpx128 zgetb(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgetb(vbase, indx);
	return ret;
}

static aplcmpx128 zgets(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgets(vbase, indx);
	return ret;
}

static aplcmpx128 zgeti(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgeti(vbase, indx);
	return ret;
}

static aplcmpx128 zgetl(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgetl(vbase, indx);
	return ret;
}

static aplcmpx128 zgetd(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgetd(vbase, indx);
	return ret;
}

static aplcmpx128 zgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initcmpx(array))(array->shape, 0);
}

static aplcmpx128 zgetz(void *vbase, BOUND indx)
{
	aplcmpx128 *base = vbase;
	aplcmpx128 ret;

	memcpy(&ret, base + indx, sizeof(ret));	// alignment ?
	return ret;
}

static aplcmpx128 zgetD(void *vbase, BOUND indx)
{
	aplcmpx128 ret;

	ret.im = 0.0;
	ret.re = dgetD(vbase, indx);
	return ret;
}

static apldecf128 Dgetb(void *vbase, BOUND indx)
{
	return igetb(vbase, indx) ? DECF_ONE : DECF_ZERO;
}

static apldecf128 Dgets(void *vbase, BOUND indx)
{
	return decf_from_int32(lgets(vbase, indx));
}

static apldecf128 Dgeti(void *vbase, BOUND indx)
{
	return decf_from_int32(lgeti(vbase, indx));
}

static apldecf128 Dgetl(void *vbase, BOUND indx)
{
	return decf_from_int32(lgetl(vbase, indx));
}

static apldecf128 Dgetd(void *vbase, BOUND indx)
{
	return decf_from_double(dgetd(vbase, indx));
}

static apldecf128 Dgetp(void *vbase, BOUND indx)
{
	ARRAY **base = vbase;
	ARRAY *array = base[indx];

	if			// subarray should be simple scalar
		(
		array->eltype == APLPNTR
	||	array->rank
		) domain();

	return (*initdecf(array))(array->shape, 0);
}

static apldecf128 Dgetz(void *vbase, BOUND indx)
{
	aplcmpx128 z = zgetz(vbase, indx);

	if(z.im)
		domain();	// strict
	return decf_from_double(z.re);
}

static apldecf128 DgetD(void *vbase, BOUND indx)
{
	apldecf128 *base = vbase;
	apldecf128 ret;

	memcpy(&ret, base + indx, sizeof(ret));
	return ret;
}

static apldecf128 Ddomx(void *vbase, BOUND indx)
	{
	domain();
	return DECF_ZERO;
	}

//	S E R V I C E S		

static aplchar8 (*charfns[])(void *vbase, BOUND indx) = 
{
			cgetc,
	cdomx,				// APLBOOL 
	cdomx,				// APLSINT 
	cdomx,				// APLINTG 
	cdomx,				// APLLONG 
	cdomx,				// APLDOUB 
			cgetp,
	cdomx,				// APLWCHAR8
	cdomx,				// APLWCHAR16
	cdomx,				// APLWCHAR32
	cdomx,				// APLCMPX
	cdomx,				// APLDECF dpd
	cdomx				// APLDECF bid
};

static aplint16 (*shortfns[])(void *vbase, BOUND indx) =
{
	idomx,				// APLNCHAR	
	igetb, igets, igeti, igetl, igetd, igetp,
	idomx, idomx, idomx,		// APLWCHARxx
	igetz, igetD, idomx
};

static aplint32 (*longfns[])(void *vbase, BOUND indx) =
{
	ldomx,				// APLNCHAR	
	lgetb, lgets, lgeti, lgetl, lgetd, lgetp,
	ldomx, ldomx, ldomx,		// APLWCHARxx
	lgetz, lgetD, ldomx
};

static aplfloat64 (*doublefns[])(void *vbase, BOUND indx) = 
{
 	ddomx,				// APLNCHAR 
	dgetb, dgets, dgeti, dgetl, getd, dgetp,
	ddomx, ddomx, ddomx,		// APLWCHARxx
	dgetz, dgetD, ddomx
};

static ARRAY	*(*pntrfns[])(void *vbase, BOUND indx) = 
{
	pgetc, pgetb, pgets, pgeti, pgetl, pgetd, pgetp,
	pgetW8, pgetW16, pgetW32,
	pgetz, pgetD, pdomx
};

static aplWchar8 (*Wchar8fns[])(void *vbase, BOUND indx) = 
{
	W8domx,
	W8domx,				// APLBOOL 
	W8domx,				// APLSINT 
	W8domx,				// APLINTG 
	W8domx,				// APLLONG 
	W8domx,				// APLDOUB 
			W8getp,		// APLPNTR
			W8getW8,	// APLWCHAR8
			W8getW16,	// APLWCHAR16
			W8getW32,	// APLWCHAR32
	W8domx,				// APLCMPX 
	W8domx,				// APLDECF 	dpd
	W8domx				// APLDECF	bid
	
};

static aplWchar16 (*Wchar16fns[])(void *vbase, BOUND indx) = 
{
	W16domx,
	W16domx,			// APLBOOL 
	W16domx,			// APLSINT 
	W16domx,			// APLINTG 
	W16domx,			// APLLONG 
	W16domx,			// APLDOUB 
			W16getp,	// APLPNTR
			W16getW8,	// APLWCHAR8
			W16getW16,	// APLWCHAR16
			W16getW32,	// APLWCHAR32
	W16domx,			// APLCMPX 
	W16domx,			// APLDECF 	dpd
	W16domx				// APLDECF	bid
	
};

static aplWchar32 (*Wchar32fns[])(void *vbase, BOUND indx) = 
{
	W32domx,
	W32domx,			// APLBOOL 
	W32domx,			// APLSINT 
	W32domx,			// APLINTG 
	W32domx,			// APLLONG 
	W32domx,			// APLDOUB 
			W32getp,	// APLPNTR
			W32getW8,	// APLWCHAR8
			W32getW16,	// APLWCHAR16
			W32getW32,	// APLWCHAR32
	W32domx,			// APLCMPX 
	W32domx,			// APLDECF 	dpd
	W32domx				// APLDECF	bid
	
};

static aplcmpx128 (*cmpxfns[])(void *vbase, BOUND indx) = 
{
 	zdomx,				// APLNCHAR 
	zgetb, zgets, zgeti, zgetl, zgetd, zgetp,
	zdomx, zdomx, zdomx,		// APLWCHARxx
	zgetz, zgetD, zdomx
};

static apldecf128 (*decffns[])(void *vbase, BOUND indx) = 
{
 	Ddomx,				// APLNCHAR 
	Dgetb, Dgets, Dgeti, Dgetl, Dgetd, Dgetp,
	Ddomx, Ddomx, Ddomx,		// APLWCHARxx
	Dgetz, DgetD, Ddomx
};

aplchar8 (*initchar(ARRAY *array))(void *vbase, BOUND indx)
{
	return charfns[array->eltype];
}

aplint16 (*initintg(ARRAY *array))(void *vbase, BOUND indx)
{
	return shortfns[array->eltype];
}

aplint32 (*initdint(ARRAY *array))(void *vbase, BOUND indx)
{
	return (sizeof(int) == sizeof(short)
			? (int (**)(void *, BOUND))shortfns
			: (int (**)(void *, BOUND))longfns)[array->eltype];
}

aplint32 (*initlong(ARRAY *array))(void *vbase, BOUND indx)
{
	return longfns[array->eltype];
}

aplfloat64 (*initdoub(ARRAY *array))(void *vbase, BOUND indx)
{
	return doublefns[array->eltype];
}

ARRAY	*(*initpntr(ARRAY *array))(void *vbase, BOUND indx)
{
	return pntrfns[array->eltype];
}

aplWchar8	(*initWchar8(ARRAY *array))(void *vbase, BOUND indx)
{
	return Wchar8fns[array->eltype];
}

aplWchar16	(*initWchar16(ARRAY *array))(void *vbase, BOUND indx)
{
	return Wchar16fns[array->eltype];
}

aplWchar32	(*initWchar32(ARRAY *array))(void *vbase, BOUND indx)
{
	return Wchar32fns[array->eltype];
}

aplcmpx128	(*initcmpx(ARRAY *array))(void *vbase, BOUND indx)
{
	return cmpxfns[array->eltype];
}

apldecf128	(*initdecf(ARRAY *array))(void *vbase, BOUND indx)
{
	return decffns[array->eltype];
}

static aplint32 roundi(double fval)	// round to integer
	{
	static double max = 0x7fffffff;
	static double min = (double)(-1) - (double) 0x7fffffff;

	fval += (fval<0 ? -.5:.5);
	if(min > fval || max < fval)
		domain();
	return fval;
	}

//	Copyright (c) 1982 Dyadic Systems Limited	
