//	Copyright (c) 1982 Dyadic Systems Limited	

#include <machine.h>

#if APL_H_INCLUDED
#else
#define APL_H_INCLUDED 1

#include <setjmp.h>
#include <signal.h>

#if	defined(DOS)
#include <dos_comms.h>
#elif	defined(DOSWIN)
#	if	defined(__NT__)
		extern void * parentid;			// HANDLE
#	else
		extern unsigned short parentid;		// HANDLE
#	endif
#	define INPIPE	((unsigned)(IPTR)parentid)
#	define OUTPIPE	((unsigned)(IPTR)parentid)
#endif	// DOS 

#ifdef INPIPE                  // eg apollo wants to override these
#define ARGSPIPE INPIPE
#define RSLTPIPE OUTPIPE
#else
#define ARGSPIPE 3             // pipe allocations
#define RSLTPIPE 4
#endif // INPIPE

#if	64 == xxBIT
#define PNTRlog		6	// 64 bit pointers		
#elif	32 == xxBIT
#define PNTRlog		5	// 32 bit pointers		
#endif
				// function syntax			
#define	RSLT	0040		// returns a result			
#define	SHYR	0140		// returns a shy result			
#define	NILA	0000		// takes no arguments			
#define	MONA	0004		// takes only a right argument		
#define	DYAD	0014		// takes a left and right argument	
#define	OPTL	0034		// left argument is optional (ambivalent)
#define	INTR	0200		// handles its own interrupts		
#define	XWID	0400		// request to return apl's X window ID  

#define	UNSMASK	(sizeof(MEMORYUNS)-1)
#define	UNSEXTRA(bytes)	((1 + ((bytes) ^ UNSMASK)) & UNSMASK)
#define	UNSROUND(bytes)	((bytes) + UNSEXTRA(bytes))
#define UNSUNITS(bytes)	((UNSROUND(bytes))/sizeof(MEMORYUNS))
#define	BITMASK	(BITSPERHEADERWORD-1)
#define	BITEXTRA(bits)	((1 + ((bits) ^ BITMASK)) & BITMASK)

#if	64 == xxBIT
#		define ALIGNWD(A)					\
			ZONESBITFIELD : (BITEXTRA(A)) - 32;		\
			ZONESBITFIELD : 16;				\
			ZONESBITFIELD : 16

#elif	32 == xxBIT
#	define ALIGNWD(A) ZONESBITFIELD         : (BITEXTRA(A))
#endif//xxBIT

#if	defined(DALIGN)
#	define		DOUBWORDS	(sizeof(double)/sizeof(MEMORYINT))
#endif	// DALIGN 

typedef struct 			// an APL ARRAY				
{

	PLENGTH       length;	// total length in words including self	
	ZONESBITFIELD type:4;   // SIMPLE|NESTED
	ZONESBITFIELD rank:4;   // rank of array
	ZONESBITFIELD eltype:4; // type of all the elements (see below)
	ZONESBITFIELD spare:4;  // to make sure that the following REFCOUNT
				// field  is 8 bit aligned
	ZONESBITFIELD refcount:8;// define a refcount field for macros that
				// think they need one
				// note that the refcount field is ignored in an
				// AP. It is only present to allow
				// the use of internal dyalog macros etc on AP ARRAYs
				// in bridge*.dll etc
	ALIGNWD(24);            // word align

	BOUND         shape[1];	
} ARRAY;

#define SIMPLE 0xf		// APL ARRAY types			
#define NESTED 0x7

#include <eltypes.h>

#if HAS_UNICODE
       #define APLCHAR8                APLWCHAR8
       #define APLACHAR                APLWCHAR32
       #if UNIX
               #define APLGCHAR        APLWCHAR32
       #else
               #define APLGCHAR        APLWCHAR16
       #endif
#else
       #define APLCHAR8                APLNCHAR
       #define APLACHAR                APLNCHAR
       #define APLGCHAR                APLNCHAR
#endif

#define	CHARlog		3
#define	BOOLlog		0
#define	SINTlog		3
#define	INTGlog		4
#define	LONGlog		5
#define	DOUBlog		6
#if	64 == xxBIT
#define PNTRlog		6	// 64 bit pointers		
#elif	32 == xxBIT
#define PNTRlog		5	// 32 bit pointers		
#endif
#define	WCHAR8log	3
#define	WCHAR16log	4
#define	WCHAR32log	5

#define ELTYPE_CHAR	0x01
#define ELTYPE_NUMB	0x02
#define ELTYPE_PNTR	0x04
#define ELTYPE_DOUB	0x10

#define ELTYPE_TYPE_MASK 	0x0f

typedef struct

	{
	ubyte type;
	ubyte size;
	ubyte shift;
	ubyte shfti;
	} ELTYPE_ATTRIBS;

static ELTYPE_ATTRIBS ELTYPE_PROPS[NELTYPES]=
	{	
/*                   TYPE	              SIZE	          SHIFT	      SHFTI             */
/*                   |                        |                   |           |	                */
/*                   |                        |                   |           |	                */
/* APLNCHAR	*/ { ELTYPE_CHAR            , sizeof(aplchar8)  , CHARlog   , CHARlog   -3	},
/* APLBOOL	*/ { ELTYPE_NUMB            , 0                 , BOOLlog   , 0			},
/* APLSINT	*/ { ELTYPE_NUMB            , sizeof(aplint8 )  , SINTlog   , SINTlog   -3	},
/* APLINTG	*/ { ELTYPE_NUMB            , sizeof(aplint16)  , INTGlog   , INTGlog   -3	},
/* APLLONG	*/ { ELTYPE_NUMB            , sizeof(aplint32)  , LONGlog   , LONGlog   -3	},
/* APLDOUB	*/ { ELTYPE_NUMB|ELTYPE_DOUB, sizeof(aplfloat64), DOUBlog   , DOUBlog   -3	},
/* APLPNTR	*/ { ELTYPE_PNTR            , sizeof(void   *)  , PNTRlog   , PNTRlog   -3	},
/* APLWCHAR8	*/ { ELTYPE_CHAR            , sizeof(aplWchar8) , WCHAR8log , WCHAR8log -3	},
/* APLWCHAR16	*/ { ELTYPE_CHAR            , sizeof(aplWchar16), WCHAR16log, WCHAR16log-3	},
/* APLWCHAR32	*/ { ELTYPE_CHAR            , sizeof(aplWchar32), WCHAR32log, WCHAR32log-3	},

	};

#define ELTYPE_TYPES(e)		(                   ELTYPE_PROPS[e].type )
#define ELTYPE_SIZE(e)		(                   ELTYPE_PROPS[e].size )
#define ELTYPE_SHIFT(e)		(                   ELTYPE_PROPS[e].shift)
#define ELTYPE_SHFTI(e)		(                   ELTYPE_PROPS[e].shfti)

#define ELTYPE_TYPE(e)		(ELTYPE_TYPE_MASK & ELTYPE_TYPES(e) )
#define ELTYPE_ISCHAR(e)	(ELTYPE_CHAR      & ELTYPE_TYPES(e))
#define ELTYPE_ISNUMB(e)	(ELTYPE_NUMB      & ELTYPE_TYPES(e))
#define ELTYPE_ISDOUB(e)	(ELTYPE_DOUB      & ELTYPE_TYPES(e))
//#define ELTYPE_ISPNTR(e)	(ELTYPE_PNTR      & ELTYPE_TYPES(e))
#define ELTYPE_ISPNTR(e)	(APLPNTR == e)


#define ARRAY_ISNUMB(p)		((p)->type==SIMPLE && ELTYPE_ISNUMB((p)->eltype))
#define ARRAY_ISCHAR(p)		((p)->type==SIMPLE && ELTYPE_ISCHAR((p)->eltype))
#define ARRAY_ISPNTR(p)		((p)->type==NESTED                              )


#define	WSFULL		1	// APL error codes			
#define	SYNTAX		2
#define	INDEX		3
#define	RANK		4
#define	LENGTH		5
#define	VALUE		6
#define	LIMIT		10
#define	EDOMAIN		11	// DOMAIN is defined in math.h
#define	NONCE		16
#define	NOPIPES		30
#define	NOPROCS		76
#define AP_EXIT		77
#define	FERR		100
#define	INTERRUPT	1003
#define	EOF_INTERRUPT	1005

				// NB error codes 200-500 are reserved	
				// for distributed auxiliary processors	
#define	PRE_ERRBASE	200	// 		prefect			
#define	VIA_ERRBASE	220	// 		via			
#define	XUT_ERRBASE	240	// 		xutils			
#define	QFS_ERRBASE	260	//		qfsck			
#define CGI_ERRBASE	280	//	cgi - taking next 53 errors	
#define NFL_ERRBASE	360	//	nfiles - taking next 20 errors	
#define PRT_ERRBASE	400	//	prt - taking the next 30 errors	

extern void domain(void);
extern void nonce(void);
extern double getd(void *vbase, BOUND indx);
extern void putd(double val, double *vbase, BOUND indx);

extern aplchar8		(*initchar   (ARRAY *))(void *vbase, BOUND indx);
extern aplint16		(*initintg   (ARRAY *))(void *vbase, BOUND indx);
extern aplint32		(*initdint   (ARRAY *))(void *vbase, BOUND indx);
extern aplint32		(*initlong   (ARRAY *))(void *vbase, BOUND indx);
extern aplfloat64	(*initdoub   (ARRAY *))(void *vbase, BOUND indx);
extern ARRAY*		(*initpntr   (ARRAY *))(void *vbase, BOUND indx);
extern aplWchar8	(*initWchar8 (ARRAY *))(void *vbase, BOUND indx);
extern aplWchar16	(*initWchar16(ARRAY *))(void *vbase, BOUND indx);
extern aplWchar32	(*initWchar32(ARRAY *))(void *vbase, BOUND indx);
extern aplcmpx128	(*initcmpx   (ARRAY *))(void *vbase, BOUND indx);
extern apldecf128	(*initdecf   (ARRAY *))(void *vbase, BOUND indx);

extern ARRAY *(*convfns[NELTYPES][NELTYPES])(ARRAY *);	// array conversion functions

#endif //APL_H_INCLUDED 1

//	Copyright (c) 1982 Dyadic Systems Limited	

